import copy
import logging
import sys
from collections import deque
from contextlib import contextmanager
from typing import Any, Sequence, Union

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone
from google.cloud.logging_v2.handlers.structured_log import StructuredLogHandler

FORMAT = "%(message)s"
IMPORT_FORMAT = "[%(namespace)s] tenant:%(tenant_name)s| %(message)s"
UNSTRUCTURED_FORMAT = "%(asctime)s |%(threadName)-10s| %(filename)31s:%(lineno)-4d |%(levelname)7s| {message_format}"

LOGGING_CONTEXT_MANAGERS = dict()  # type: dict[str, LoggingContextHandler]


def get_logger(name: str) -> Any:
    if name not in LOGGING_CONTEXT_MANAGERS:
        logging_context_handler = LoggingContextHandler()
        context_filter = ImportContextFilter(logging_context_handler)

        logger = setup_logging(name=name, format=IMPORT_FORMAT)
        logger.addFilter(context_filter)

        LOGGING_CONTEXT_MANAGERS[name] = logging_context_handler
        return logger

    logger = logging.getLogger(name)
    return logger


class LoggingContextHandler:
    """Allows having a stack of logging context variables.
    Each time you add to the stack the logger will get new context variables.
    Each time you remove it will go to the ones it had previously.
    This allows stacking like:
    with logger_context(contextvars_1):
        log.info() -> log with contextvars_1

        with logger_context(contextvars_2):
            log.info() -> log with contextvars_2

        log.info() -> log with contextvars_1 again
    """

    def __init__(self):
        self.attributes = deque([{}])  # type: deque[dict]

    @property
    def current_attributes(self) -> dict:
        return self.attributes[0]

    def add(self, **new_context_vars) -> None:
        old_context = self.attributes[0]
        new_context = {**old_context, **new_context_vars}
        self.attributes.appendleft(new_context)

    def get(self, key: str, default: Any = None) -> Any:
        return self.attributes[0].get(key, default)

    def remove(self) -> None:
        self.attributes.popleft()

    def __str__(self) -> str:
        return str(self.attributes)


class ImportContextFilter(logging.Filter):
    def __init__(self, import_logging_context_handler: LoggingContextHandler):
        super(ImportContextFilter, self).__init__()
        self.import_logging_context_handler = import_logging_context_handler

    def filter(self, record: logging.LogRecord) -> bool:
        record.tenant_name = self.import_logging_context_handler.get("tenant_name")

        namespace_str = ">".join(
            self.import_logging_context_handler.get("namespace", [])
        )
        record.namespace = namespace_str.upper()
        record.json_fields = {
            **getattr(record, "json_fields", {}),
            "tenant_name": record.tenant_name,
            "namespace": record.namespace,
        }
        return True


class TimestampFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        record.json_fields = {
            **getattr(record, "json_fields", {}),
            "timestamp": str(timezone.now()),
        }
        return True


@contextmanager
def append_logger_context_namespace(logger: logging.Logger, namespace: str) -> None:
    logging_context_handler = LOGGING_CONTEXT_MANAGERS[logger.name]
    try:
        next_attributes = copy.deepcopy(logging_context_handler.current_attributes)
        current_namespace = next_attributes.get("namespace", [])
        next_attributes["namespace"] = current_namespace + [namespace]
        logging_context_handler.add(**next_attributes)
        yield
    finally:
        logging_context_handler.remove()


@contextmanager
def logger_context(
    logger: logging.Logger, namespace: Union[str, Sequence[str]], **kwargs
) -> None:
    logging_context_handler = LOGGING_CONTEXT_MANAGERS[logger.name]
    try:
        if isinstance(namespace, str):
            # convert namespace to sequence
            namespace = [namespace]
        logging_context_handler.add(namespace=namespace, **kwargs)
        yield
    finally:
        logging_context_handler.remove()


def _structured_log() -> bool:
    try:
        return settings.STRUCTURED_LOG
    except (ImproperlyConfigured, AttributeError):
        return False


def _get_handler(format: str) -> logging.StreamHandler:
    if _structured_log():
        log_handler: logging.StreamHandler = StructuredLogHandler(stream=sys.stderr)
        log_handler.setFormatter(logging.Formatter(FORMAT))
    else:
        log_handler = logging.StreamHandler(stream=sys.stderr)
        log_handler.setFormatter(
            logging.Formatter(UNSTRUCTURED_FORMAT.format(message_format=format))
        )
    return log_handler


def setup_logging(name: str, format: str) -> logging.Logger:
    """Sets up and returns the recommendation logger."""
    logger = logging.getLogger(name)
    logger.setLevel(
        logging.DEBUG
    )  # logger passes on everything, but the handler may filter
    logger.handlers = []
    log_handler = _get_handler(format)
    log_handler.setLevel(logging.INFO)
    logger.addHandler(log_handler)

    if _structured_log():
        logger.addFilter(TimestampFilter())

    return logger



import json
import socket
import sys
from functools import wraps
from typing import Iterator, Optional, Tuple

from . import helpers
from . import models
from .settings import HOST, PORT


class Leek(object):
    def task(self, f, pool=None):
        pool_name = pool or f.__name__

        @wraps(f)
        def _offload(*args, **kwargs):
            return push_task_to_queue(f, pool_name=pool_name, *args, **kwargs)

        f.offload = _offload
        return f


class Task(object):
    def __init__(self, a_callable, *args, **kwargs):
        assert callable(a_callable)
        self.task_callable = a_callable
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.task_callable(*self.args, **self.kwargs)


def start_task_with_id(task_id: int):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((HOST, PORT))
    sock.send("{}".format(task_id).encode())
    received = sock.recv(1024)
    sock.close()
    return json.loads(received.decode())


def push_task_to_queue(a_callable, *args, **kwargs):
    """Original API"""
    pool_name = kwargs.pop('pool_name', None)

    new_task = Task(a_callable, *args, **kwargs)
    queued_task = helpers.save_task_to_db(new_task, pool_name)

    return start_task_with_id(queued_task.id)


def query_task(task_id: int) -> models.Task:
    return helpers.load_task(task_id)


def list_tasks(finished: Optional[bool] = None) -> Iterator[Tuple[models.Task, Task]]:
    db_tasks = models.Task.objects.all().order_by('queued_at')
    if finished is not None:
        db_tasks = db_tasks.filter(finished_at__isnull=not finished)
    for db_task in db_tasks:
        try:
            task = helpers.unpack(db_task.pickled_task)
            yield db_task, task
        except (ModuleNotFoundError, AttributeError):  # things that can happen during unpickle
            print("could not unpickle task", db_task.id, file=sys.stderr)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2020-03-10 09:18
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_leek', '0003_auto_20180910_1028'),
    ]

    operations = [
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('pickled_task', models.BinaryField(max_length=4096)),
                ('pool', models.CharField(max_length=256, null=True)),
                ('queued_at', models.DateTimeField(auto_now_add=True)),
                ('started_at', models.DateTimeField(null=True)),
                ('finished_at', models.DateTimeField(null=True)),
                ('pickled_exception', models.BinaryField(max_length=2048, null=True)),
                ('pickled_return', models.BinaryField(max_length=4096, null=True)),
            ],
        ),
        migrations.DeleteModel(
            name='FailedTasks',
        ),
        migrations.DeleteModel(
            name='QueuedTasks',
        ),
        migrations.DeleteModel(
            name='SuccessTasks',
        ),
    ]



import base64
import pickle

from . import models


def unpack(pickled_task: bytes):
    new_task = pickle.loads(base64.b64decode(pickled_task))
    return new_task


def serialize(obj) -> bytes:
    return base64.b64encode(pickle.dumps(obj))


def serialize_exception(e: Exception) -> bytes:
    try:
        return serialize(e)
    except Exception:
        return serialize(Exception("Failed to serialize exception: %s" % e))


def load_task(task_id: int) -> models.Task:
    return models.Task.objects.get(pk=task_id)


def save_task_to_db(new_task, pool_name: str) -> models.Task:
    pickled_task = serialize(new_task)
    t = models.Task(pickled_task=pickled_task, pool=pool_name)
    t.save()
    return t



import json
import multiprocessing
import queue
import socketserver
import threading
from sys import platform

import django
from django.utils import timezone

from .cloud_logging import get_logger, logger_context
from . import helpers
from .helpers import load_task

log = get_logger(__name__)


def target(queue):
    django.setup()
    log.info("Worker Starts")
    while True:
        task_id = queue.get()
        if task_id is None:
            return

        log.info("loading task...")

        # workaround to solve problems with django + psycopg2
        # solution found here: https://stackoverflow.com/a/36580629/10385696
        django.db.connection.close()

        task = load_task(task_id=task_id)
        if task.finished():
            log.info("skipping %s as it is already finished", task_id)
            continue

        pickled_task = helpers.unpack(task.pickled_task)

        with logger_context(logger=log, namespace=task.pool, **pickled_task.kwargs):
            log.info("running task...")
            try:
                task.started_at = timezone.now()
                task.save()
                return_value = pickled_task()
                task.finished_at = timezone.now()
                task.pickled_return = helpers.serialize(return_value)
                task.save()

                log.info("...successfully")
            except Exception as e:
                log.exception("...task failed")
                task.finished_at = timezone.now()
                task.pickled_exception = helpers.serialize_exception(e)
                task.save()


class Pool(object):
    def __init__(self):
        if platform == "darwin":
            # OSX does not support forking
            self.queue = queue.Queue()
            self.worker = threading.Thread(target=target, args=(self.queue,))
        else:
            self.queue = multiprocessing.Queue()
            self.worker = multiprocessing.Process(target=target, args=(self.queue,))

    def stop(self):
        self.queue.put(None)


class TaskSocketServer(socketserver.BaseRequestHandler):
    DEFAULT_POOL = "default"
    # pools holds a mapping from pool names to process objects
    pools = {}

    def handle(self):
        try:
            data = self.request.recv(5000).strip()

            # assume a serialized task
            log.info("Got a task")
            task_id = None
            try:
                task_id = int(data.decode())

                # Connection are closed by tasks, force it to reconnect
                django.db.connections.close_all()
                task = load_task(task_id=task_id)

                # Ensure pool got a worker processing it
                pool_name = task.pool or self.DEFAULT_POOL
                pool = self.pools.get(pool_name)
                if pool is None or not pool.worker.is_alive():
                    # Spawn new pool
                    log.info("Spawning new pool: {}".format(pool_name))
                    self.pools[pool_name] = Pool()
                    self.pools[pool_name].worker.start()

                self.pools[pool_name].queue.put(task_id)

                response = {"task": "queued", "task_id": task_id}
            except Exception as e:
                log.exception("failed to queue task")
                response = {
                    "task": "failed to queue",
                    "task_id": task_id,
                    "error": str(e),
                }

            self.request.send(json.dumps(response).encode())

        except OSError as e:
            # in case of network error, just log
            log.exception("network error")

    @staticmethod
    def stop():
        for name, pool in TaskSocketServer.pools.items():
            log.info("Stopping pool: %s", name)
            pool.stop()



import base64
import pickle
from typing import Any

from django.db import models


class Task(models.Model):
    pickled_task = models.BinaryField(max_length=4096)
    pool = models.CharField(max_length=256, null=True)
    queued_at = models.DateTimeField(auto_now_add=True)
    started_at = models.DateTimeField(null=True)
    finished_at = models.DateTimeField(null=True)
    pickled_exception = models.BinaryField(max_length=2048, null=True)
    pickled_return = models.BinaryField(max_length=4096, null=True)

    @property
    def exception(self):
        if self.pickled_exception is None:
            return None
        return pickle.loads(base64.b64decode(self.pickled_exception))

    @property
    def return_value(self):
        if self.pickled_return is None:
            return None
        return pickle.loads(base64.b64decode(self.pickled_return))

    def started(self) -> bool:
        return self.started_at is not None

    def finished(self) -> bool:
        return self.finished_at is not None

    def successful(self) -> bool:
        return self.finished() and self.pickled_return is not None



import os
from setuptools import find_packages, setup

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


with open("README.md") as f:
    README = f.read()


setup(
    name="django-leek",
    version="1.0.6",
    packages=find_packages(exclude=["test_app"]),
    install_requires=["django>=3.2", "google-cloud-logging>=3.0.0"],
    include_package_data=True,
    license="MIT License",
    description="A simple Django app to offload tasks from main web server",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Volumental/django-leek",
    author="Volumental",
    author_email="maintainer@volumental.com",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)
